package registry

import (
	"fmt"
	"reflect"
)

// Global

// Description of an entry
type Description struct {
	Name     string
	Category []string
	Inputs   []string
	Output   string
}

// Entry contains a function description params etc
type Entry struct {
	fn interface{}
}

// Registry function registry
type Registry struct {
	data map[string]*Entry
}

// New creates a new registry
func New() *Registry {
	return &Registry{map[string]*Entry{}}
}

//Register should be a func only register
func (r *Registry) Register(name string, v interface{}) error {
	fntyp := reflect.TypeOf(v)
	if fntyp.Kind() != reflect.Func {
		return ErrNotAFunc
	}
	if fntyp.NumOut() == 0 {
		return ErrOutput
	}

	r.data[name] = &Entry{fn: v}
	return nil
}

// Get an entry
func (r *Registry) Get(name string, params ...interface{}) (interface{}, error) {
	e, ok := r.data[name]
	if !ok {
		return nil, fmt.Errorf("Entry '%s' not found", name)
	}
	v := e.fn
	// We already know this is a function
	// and that returns 1 or more values
	vtyp := reflect.TypeOf(v)
	if vtyp.Out(0).Kind() == reflect.Func {
		fparam := make([]reflect.Value, len(params))
		for i := range params {
			fparam[i] = reflect.ValueOf(params[i])
		}
		// Call the func and return the thing
		v = reflect.ValueOf(v).Call(fparam)[0].Interface()
	}

	return v, nil
}

//Describe named fn
func (r *Registry) Describe(name string) (*Description, error) {
	e, ok := r.data[name]
	if !ok {
		return nil, fmt.Errorf("Entry '%s' not found", name)
	}
	fnTyp := reflect.TypeOf(e.fn)
	Inputs := []string{}

	nInputs := fnTyp.NumIn()
	for i := 0; i < nInputs; i++ {
		Inputs = append(Inputs, fmt.Sprint(fnTyp.In(i)))
	}
	Output := fmt.Sprint(fnTyp.Out(0))
	desc := &Description{
		Name:     name,
		Category: []string{"generic"},
		Inputs:   Inputs,
		Output:   Output,
	}

	return desc, nil
}

// Descriptions Description list
func (r *Registry) Descriptions() ([]Description, error) {
	ret := make([]Description, len(r.data))
	i := 0
	for k := range r.data {
		d, err := r.Describe(k)
		if err != nil {
			return nil, err
		}
		ret[i] = *d
		i++
	}
	return ret, nil
}
